#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define AIN      (1<<PA0)
#define RESET_   (1<<PA2)
#define CS_      (1<<PA3)
#define SCL      (1<<PA4)
#define SI       (1<<PA5)
#define A0       (1<<PA7)

#define DRIVE_H1 (1<<PB0)
#define DRIVE_L1 (1<<PB1)
#define DRIVE_H2 (1<<PA1)
#define DRIVE_L2 (1<<PA6)
#define BUTTON   (1<<PB2)

#define EEPROM_ADDR_ZERO_LOW    (uint32_t*)0
#define EEPROM_ADDR_ZERO_HIGH   (uint32_t*)4
#define EEPROM_ADDR_VALUE_LOW   (uint32_t*)8
#define EEPROM_ADDR_VALUE_HIGH  (uint32_t*)12

#define CAL_RESISTANCE_LOW  50.0
#define CAL_RESISTANCE_HIGH 1000.0

// first byte: width, second byte: lines
// the following (width * lines) bytes specify the row data
// Gimp font: Arial Bold 22pt

#define MAX_CHAR_WIDTH 17
#define MAX_CHAR_LINES 3

typedef struct
{
    uint8_t rows;
    uint8_t columns;
    uint8_t data[MAX_CHAR_LINES][MAX_CHAR_WIDTH];
} char_def;

const __flash char_def const characters[] =
{
    // 0
    {
        2, 10,
        {
            {0x1f, 0x7f, 0x7f, 0xe0, 0xc0, 0xc0, 0xe0, 0x7f, 0x7f, 0x1f},
            {0xf8, 0xfe, 0xfe, 0x07, 0x03, 0x03, 0x07, 0xfe, 0xfe, 0xf8}
        }
    },
    // 1
    {
        2, 7,
        {
            {0x0e, 0x1c, 0x1c, 0x38, 0xff, 0xff, 0xff},
            {0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff}
        }
    },
    // 2
    {
        2, 10,
        {
            {0x30, 0x70, 0xf0, 0xe0, 0xc0, 0xc1, 0xe3, 0xff, 0x7f, 0x3c},
            {0x03, 0x0f, 0x1f, 0x3f, 0xfb, 0xf3, 0xc3, 0x83, 0x03, 0x03}
        }
    },
    // 3
    {
        2, 10,
        {
            {0x10, 0x70, 0x70, 0xe0, 0xc3, 0xc7, 0xff, 0x7f, 0x3c, 0x00},
            {0x0c, 0x0e, 0x0f, 0x07, 0x03, 0x03, 0x87, 0xfe, 0xfe, 0x78}
        }
    },
    // 4
    {
        2, 11,
        {{0x00, 0x00, 0x01, 0x07, 0x1e, 0x38, 0xff, 0xff, 0xff, 0x00, 0x00},
         {0x38, 0xf8, 0xd8, 0x98, 0x18, 0x18, 0xff, 0xff, 0xff, 0x18, 0x18}}
    },
    // 5
    {
        2, 10,
        {{0x03, 0x1f, 0xff, 0xfe, 0xe6, 0xc6, 0xc7, 0xc7, 0xc3, 0x00},
         {0x0c, 0x0e, 0x0f, 0x07, 0x03, 0x03, 0x07, 0xfe, 0xfc, 0xf8}}
    },
    // 6
    {
        2, 10,
        {{0x1f, 0x7f, 0x7f, 0xe1, 0xc3, 0xc3, 0xe3, 0xf3, 0x71, 0x30},
         {0xf8, 0xfe, 0xfe, 0x87, 0x03, 0x03, 0x87, 0xfe, 0xfe, 0x78}}
    },
    // 7
    {
        2, 9,
        {{0xc0, 0xc0, 0xc0, 0xc0, 0xc3, 0xcf, 0xdf, 0xfc, 0xe0},
         {0x00, 0x00, 0x0f, 0xff, 0xff, 0xf0, 0x00, 0x00, 0x00}}
    },
    // 8
    {
        2, 10,
        {{0x3c, 0x7e, 0x7f, 0xe3, 0xc1, 0xc1, 0xe3, 0x7f, 0x7e, 0x3c},
         {0x3c, 0x7e, 0xfe, 0xc7, 0x83, 0x83, 0xc7, 0xfe, 0x7e, 0x3c}}
    },
    // 9
    {
        2, 10,
        {{0x1e, 0x7f, 0x7f, 0xe1, 0xc0, 0xc0, 0xe1, 0x7f, 0x7f, 0x1f},
         {0x0c, 0x8e, 0xcf, 0xc7, 0xc3, 0xc3, 0x87, 0xfe, 0xfe, 0xf8}}
    },
    // µ
    {
        3, 11,
        {{0x0f, 0x0f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0f, 0x0f, 0x0f},
         {0xff, 0xff, 0xff, 0x06, 0x03, 0x03, 0x03, 0x06, 0xff, 0xff, 0xff},
         {0xf0, 0xf0, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}}
    },
    // m
    {
        2, 17,
        {{0x0f, 0x0f, 0x0f, 0x06, 0x0c, 0x0c, 0x0c, 0x0f, 0x0f, 0x07, 0x06, 0x0c, 0x0c, 0x0c, 0x0f, 0x07, 0x03},
         {0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff}}
    },
    // Omega
    {
        2, 16,
        {{0x07, 0x1f, 0x3f, 0x7c, 0x70, 0xe0, 0xe0, 0xe0, 0xe0, 0xe0, 0xe0, 0x70, 0x7c, 0x3f, 0x1f, 0x07},
         {0x87, 0xe7, 0xf7, 0x3f, 0x1f, 0x0f, 0x07, 0x00, 0x00, 0x07, 0x0f, 0x1f, 0x3f, 0xf7, 0xe7, 0x87}}
    },
    // .
    {
        2, 3,
        {{0x00, 0x00, 0x00},
         {0x07, 0x07, 0x07}}
    },
    // -
    {
        2, 7,
        {{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
         {0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70}}
    },
    // >
    {
        2, 11,
        {{0x38, 0x1c, 0x1c, 0x1c, 0x0e, 0x0e, 0x07, 0x07, 0x03, 0x03, 0x03},
         {0x1c, 0x38, 0x38, 0x38, 0x70, 0x70, 0xe0, 0xe0, 0xc0, 0xc0, 0xc0}}
    },
    // space
    {
        2, 4
    }
};

// assumed initial state for all functions:
// - CS_ => H
// - A0, DATA => *

void lcd_put_serial(uint8_t data)
{
    for (int i = 0; i < 8; i++)
    {
        PORTA &= ~SCL;
        if (data & 0x80)
        {
            PORTA |= SI;
        }
        else
        {
            PORTA &= ~SI;
        }
        data <<= 1;
        PORTA |= SCL;
    }
}

void lcd_command(uint8_t cmd)
{
    PORTA &= ~(CS_ | A0);
    lcd_put_serial(cmd);
    PORTA |= CS_;
}

void lcd_write_data(uint8_t data)
{
    PORTA |= A0;
    PORTA &= ~CS_;
    lcd_put_serial(data);
    PORTA |= CS_;
}

inline void lcd_display_on(uint8_t on)
{
    lcd_command(0xAE | (on & 1));
}

// valid range: 0-63
void lcd_set_contrast(uint8_t value)
{
    lcd_command(0x81);
    lcd_command(value);
}

// valid range: 0-7
inline void lcd_set_power(uint8_t value)
{
    lcd_command(0x28 | (value & 0x07));
}

// valid range: 0-7
inline void lcd_set_resistor_ratio(uint8_t value)
{
    lcd_command(0x20 | (value & 0x07));
}

// valid range: 0-63
inline void lcd_set_start_line(uint8_t value)
{
    lcd_command(0x40 | (value & 0x2F));
}

// valid range: 0-63
inline void lcd_set_page(uint8_t page)
{
    lcd_command(0xB0 | (page & 0x0F));
}

// valid range: 0-132
void lcd_set_column(uint8_t col)
{
    lcd_command(0x10 | (col >> 4));
    lcd_command(col & 0x0F);
}

#define ADC_NORMAL  0
#define ADC_REVERSE 1
inline void lcd_adc_select(uint8_t value)
{
    lcd_command(0xA0 | (value & 0x01));
}

#define ON_NORMAL   0
#define ON_REVERSE  1
inline void lcd_reverse_display(uint8_t value)
{
    lcd_command(0xA6 | (value & 0x01));
}

#define NORMAL_ON   0
#define ALL_ON      1
inline void lcd_entire_display_on(uint8_t value)
{
    lcd_command(0xA4 | (value & 0x01));
}

inline void lcd_reset(void)
{
    lcd_command(0xE2);
}

inline void lcd_nop(void)
{
    lcd_command(0xE3);
}

void lcd_clear(void)
{
    int page, col;
    for (page = 0; page < 4; page++)
    {
        lcd_set_page(page);
        lcd_set_column(0);
        for (col = 0; col < 132; col++)
        {
            lcd_write_data(0);
        }
    }
}

uint8_t lcd_write_char(uint8_t c, uint8_t page, uint8_t column)
{
    const char list[] = "0123456789umO.-> ";
    char* pos = strchr(list, c);
    if (pos != NULL)
    {
        uint8_t row, col;
        char_def d;

        c = pos - list;
        memcpy_P(&d, &characters[c], sizeof(char_def));

        for (row = 0; row < d.rows; row++)
        {
            lcd_set_page(page);
            lcd_set_column(column);
            for (col = 0; col < d.columns; col++)
            {
                lcd_write_data(d.data[row][col]);
            }
            // add two empty columns
            lcd_write_data(0);
            lcd_write_data(0);
            page--;
        }

        return d.columns + 2;
    }

    return 0;
}

void lcd_write_string(char* s, uint8_t page, uint8_t column)
{
    while (*s != '\0')
    {
        column += lcd_write_char(*s++, page, column);
    }
}

void lcd_write_line(char* s)
{
    uint8_t column = 0;
    while (*s != '\0')
    {
        column += lcd_write_char(*s++, 2, column);
    }
    while (column < 132)
    {
        column += lcd_write_char(' ', 2, column);
    }
}

void display_setup(void)
{
    PORTA |= CS_;

    _delay_ms(10);
    PORTA |= RESET_;
    _delay_ms(10);

    // power control
    lcd_set_power(0x07);

    lcd_reset();

    // V0 voltage regulator internal resistor ratio set
    lcd_set_resistor_ratio(0x03);

    // electronic volume mode set
    lcd_set_contrast(0x24);
    //lcd_adc_select(ADC_REVERSE);

    lcd_clear();

    // switch display on
    lcd_display_on(1);
}

#define DELAY1 200
#define DELAY2 155

int32_t measure_range_1000mOhm(void)
{
    int16_t cycle;
    int32_t adc_sum = 0;

    // switch off DRIVE_H1 and DRIVE_L1
    PORTB &= ~(DRIVE_H1 | DRIVE_L1);
    // configure DRIVE_H1 and DRIVE_L1 as inputs
    DDRB &= ~(DRIVE_H1 | DRIVE_L1);
    // configure DRIVE_H2 and DRIVE_L2 as outputs
    DDRA |= DRIVE_H2 | DRIVE_L2;

    // bring drive pins into a defined start state
    PORTA |= DRIVE_H2;
    PORTA &= ~DRIVE_L2;

    for (cycle = 0; cycle < 2000; cycle++)
    {
        // cycle starts by toggling DRIVE_H2 and DRIVE_L2
        PORTA ^= DRIVE_H2 | DRIVE_L2;

        // wait some microseconds for the output states to stabilize
        _delay_us(DELAY1);

        // start ADC conversion
        ADCSRA |= (1<<ADSC);

        // wait for ADC conversion to complete
        _delay_us(DELAY2);

        if (cycle & 1)
        {
            adc_sum += ADC;
        }
        else
        {
            adc_sum -= ADC;
        }
    }

    return adc_sum;
}

uint8_t overrange;

#define OVERRANGE_CHECK_CYCLE 10
#define OVERRANGE_LOW  (1023L * OVERRANGE_CHECK_CYCLE)
#define OVERRANGE_HIGH (1023L * 1000)

int32_t measure_range_100mOhm(void)
{
    int16_t cycle;
    int32_t adc_sum = 0;
    overrange = 0;

    // switch off DRIVE_H2 and DRIVE_L2
    PORTA &= ~(DRIVE_H2 | DRIVE_L2);
    // configure DRIVE_H2 and DRIVE_L2 as inputs
    DDRA &= ~(DRIVE_H2 | DRIVE_L2);
    // configure DRIVE_H1 and DRIVE_L1 as outputs
    DDRB |= DRIVE_H1 | DRIVE_L1;

    // bring drive pins into a defined start state
    PORTB |= DRIVE_H1;
    PORTB &= ~DRIVE_L1;

    for (cycle = 0; cycle < 2000; cycle++)
    {
        // cycle starts by toggling DRIVE_H1 and DRIVE_L1
        PORTB ^= DRIVE_H1 | DRIVE_L1;

        // wait some microseconds for the output states to stabilize
        _delay_us(DELAY1);

        // start ADC conversion
        ADCSRA |= (1<<ADSC);

        // wait for ADC conversion to complete
        _delay_us(DELAY2);

        if (cycle & 1)
        {
            adc_sum += ADC;
        }
        else
        {
            adc_sum -= ADC;
        }

        if ((cycle == OVERRANGE_CHECK_CYCLE * 2) && (adc_sum >= OVERRANGE_LOW))
        {
            overrange = 1;
            break;
        }
    }

    return adc_sum;
}

volatile uint8_t button_clicked = 0;

ISR(PCINT1_vect)
{
    if ((PINB & BUTTON) == 0)
    {
        button_clicked = 1;
    }
}

void wait_for_button(void)
{
    while ((PINB & BUTTON) != 0)
    {
        _delay_ms(10);
    }
    while ((PINB & BUTTON) == 0)
    {
        _delay_ms(10);
    }
}

int32_t zero_low;
int32_t zero_high;
int32_t value_low;
int32_t value_high;

double scale_low;
double scale_high;

const __flash char const cal_zero[] = "--> 0 O";
const __flash char const cal_low[]  = "--> 50 mO";
const __flash char const cal_high[] = "--> 1 O";

void calibrate(void)
{
    char line[16];

    // calibrate 0 Ohm
    memcpy_P(line, cal_zero, sizeof(cal_zero));
    lcd_write_line(line);
    wait_for_button();
    zero_low = measure_range_100mOhm();
    sprintf(line, "%ld", zero_low);
    lcd_write_line(line);
    _delay_ms(2000);
    zero_high = measure_range_1000mOhm();
    sprintf(line, "%ld", zero_high);
    lcd_write_line(line);
    _delay_ms(2000);

    // calibrate 50 mOhm
    memcpy_P(line, cal_low, sizeof(cal_low));
    lcd_write_line(line);
    wait_for_button();
    value_low = measure_range_100mOhm();
    sprintf(line, "%ld", value_low);
    lcd_write_line(line);
    _delay_ms(2000);

    // calibrate 1 Ohm
    memcpy_P(line, cal_high, sizeof(cal_high));
    lcd_write_line(line);
    wait_for_button();
    value_high = measure_range_1000mOhm();
    sprintf(line, "%ld", value_high);
    lcd_write_line(line);
    _delay_ms(2000);

    eeprom_update_dword(EEPROM_ADDR_ZERO_LOW,   zero_low);
    eeprom_update_dword(EEPROM_ADDR_ZERO_HIGH,  zero_high);
    eeprom_update_dword(EEPROM_ADDR_VALUE_LOW,  value_low);
    eeprom_update_dword(EEPROM_ADDR_VALUE_HIGH, value_high);

    scale_low  = (value_low - zero_low) / CAL_RESISTANCE_LOW;
    scale_high = (value_high - zero_high) / CAL_RESISTANCE_HIGH;
}

int main(void)
{
    int32_t adc_sum;
    double value;
    char string[20];

    MCUCR |= (1<<PUD); // disable pull-ups globally
    DDRA = RESET_ | CS_ | SCL | SI | A0;
    DDRB = DRIVE_L1 | DRIVE_H1;
    PORTB |= DRIVE_H1;

    // enable ADC, divide sysclk by 8 (=> 125kHz)
    ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS0);
    DIDR0 = (1<<ADC0D);

    PCMSK1 = (1<<PCINT10);
    GIMSK = (1<<PCIE1);

    zero_low = eeprom_read_dword(EEPROM_ADDR_ZERO_LOW);
    zero_high = eeprom_read_dword(EEPROM_ADDR_ZERO_HIGH);
    value_low = eeprom_read_dword(EEPROM_ADDR_VALUE_LOW);
    value_high = eeprom_read_dword(EEPROM_ADDR_VALUE_HIGH);

    scale_low  = (value_low - zero_low) / CAL_RESISTANCE_LOW;
    scale_high = (value_high - zero_high) / CAL_RESISTANCE_HIGH;

    display_setup();

    if ((PINB & BUTTON) == 0)
    {
        while ((PINB & BUTTON) == 0)
        {
            // wait for button release
            _delay_ms(10);
        }

        calibrate();
    }

    sei();

    while (1)
    {
        adc_sum = measure_range_100mOhm();
        if (overrange)
        {
            adc_sum = measure_range_1000mOhm();
            value = (double)(adc_sum - zero_high) / scale_high;
            if (adc_sum >= OVERRANGE_HIGH)
            {
                sprintf(string, "> %-.1f mO", value);
            }
            else
            {
                sprintf(string, "%-.1f mO", value);
            }
        }
        else
        {
            value = (double)(adc_sum - zero_low) / scale_low;
            sprintf(string, "%-.2f mO", value);
        }

        //sprintf(string, "%d %ld", overrange, adc_sum);

        lcd_write_line(string);

        // check zero-set button
        if (button_clicked)
        {
            zero_low = measure_range_100mOhm();
            zero_high = measure_range_1000mOhm();
            button_clicked = 0;
        }
    }

    return 0;
}

